/**
File: DivisonManagerLM2plus.java
--------------------------------------------------------
Creates a division manager object for learning model 2+. 
*
* Author:  Iris Lorscheid
* Date: May, 1st 2014 
*/

package agents;

import java.io.FileWriter;
import java.io.IOException;
import java.util.Hashtable;

import knowledgebase.Knowledgebase;
import learning.SarinVahidRL;
import library.ArrayOperations;
import library.MathOperations;
import library.ProfitCalculation;
import observer.ObserverLM2plus;

public class DivisionManagerLM2plus extends DivisionManager {
	protected String identifier;
	protected String stateKey;
	protected ObserverLM2plus observer;
	protected Knowledgebase kb;
	
	protected SarinVahidRL la;
	protected int k, counter;
	protected Hashtable<Double, Integer> exploreCounter;
	
	/**
	 * Creates division manager agent for learning model LM1. 
	 * @param identifier - Identifier for the division manager agent
	 * @param reductProd - Productivity reduction
	 * @param optCapacity - Optimal capacity (threshold I)
	 * @param maxCapacity - Maximum capacity (threshold II)
	 * @param pMin - Minimum productivity value
	 * @param prodScaleRep - Scale of possible productivity values
	 * @param incrRep - Increment between productivity values
	 * @param verificationOutput - FileWriter to print output
	 * @param observer - Observer object
	 * @throws IOException 
	 */
	public DivisionManagerLM2plus(String identifier, double reductProd, int optCapacity, 
			int maxCapacity, double pMin, int prodScaleRep, double incrRep, 
			FileWriter verificationOutput, int k, ObserverLM2plus observer) throws IOException{
		
		super(reductProd, optCapacity,  maxCapacity,  pMin, prodScaleRep, incrRep, verificationOutput);
		this.identifier = identifier;							
		this.observer = observer;
		this.k = k;
		this.counter = 0;
		this.exploreCounter = new Hashtable<Double, Integer>(); 
		for (double c = pMin; c < (pMin + (prodScaleRep * incrRep)); c+= incrRep ){ 
			c = MathOperations.roundTo1(c);
			exploreCounter.put(c,0);
		}
		
		kb = new Knowledgebase(		ArrayOperations.generateArray(prodScaleRep, pMin, incrRep),
				 												MathOperations.roundTo4( (double) 1 / (double) prodScaleRep),
																identifier
																);
		
		setLa(new SarinVahidRL(identifier));
	}
	
	/* (non-Javadoc)
	 * @see agents.DivisionManager#reportProductivity(double)
	 */
	public double reportProductivity(double realP) throws IOException{
		prodValue = realP;
		setStateKey(realP + "#");					// LM2 - explicit experience collection with  realP
		
		if(counter == 0){
			repProd = kb.ask(stateKey);
			exploreCounter.put
					(repProd, 
					exploreCounter.get(repProd)+1);
		} else
		{
			if(exploreCounter.get(repProd) < k){
				kb.ask(stateKey);  							// Method call only to create state
				exploreCounter.put(repProd, exploreCounter.get(repProd)+1);
			} else{
				repProd = kb.ask(stateKey);
			}
		}
		counter ++;
		return MathOperations.roundTo1(repProd);
	}
	
	/* (non-Javadoc)
	 * @see agents.DivisionManager#produce(int)
	 */
	public double produce(int resources){	
		resourcesAllocated = resources;
		profit = ProfitCalculation.calcProfit(resources, prodValue, observer);
		return profit;
	}
	
	/* (non-Javadoc)
	 * @see agents.DivisionManager#receiveCompensation(double)
	 */
	public void receiveCompensation(double payment){
		compensation = payment;
		}
		
	/**
	 * Updates the knowledge base of the division manager
	 * according to feedback received 
	 *  
	 * @param feedback
	 */
	public void updateKB(double feedback){
			try {
				la.updateKB(		kb.getStates().get(stateKey), 
						Double.toString(repProd), 
						feedback);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	
	public String getIdentifier() {
		return identifier;
	}

	public String getStateKey() {
		return stateKey;
	}

	public void setStateKey(String stateKey) {
		this.stateKey = stateKey;
	}

	public SarinVahidRL getLa() {
		return la;
	}

	public void setLa(SarinVahidRL la) {
		this.la = la;
	}
	
	public Knowledgebase getKb() {
		return kb;
	}

	public void setKb(Knowledgebase kb) {
		this.kb = kb;
	}

	
}
