/** 
File: Beliefs.java
--------------------------------------------------------
Creates a belief object for ficticious play learning model. 
* 
* Author:  Iris Lorscheid
* Date: May, 1st 2014 
*/



package knowledgebase;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import library.MathOperations;


public class Beliefs  extends Knowledgebase{
	private Hashtable<String, Integer> observations;
	private String belief;

	public String getBelief() {
		return belief;
	}

	private double random;
	private double threshold;
	private double maxValue;
	public double getMaxValue() {
		return maxValue;
	}

	private ArrayList<String> predictions;

	/**
	 * Creates beliefs as part of the knowledgebase
	 * 
	 * @param actions
	 * @param initial_strength
	 * @param agent_identifier
	 */
	public Beliefs(double[] actions, double initial_strength,	String agent_identifier){
		super(actions, initial_strength, agent_identifier);

		observations = new Hashtable<String, Integer>();
		belief = new String();
		random = 0.0;
		threshold = 0.0;
		maxValue = 0.0;
		predictions = new ArrayList<String>();
	}
	
	/**
	 * Derives belief for given time t
	 * @return belief
	 */
	public String formulateBelief(int t){
		if (t == 0){
			belief = chooseBeliefStochastically();
			//System.out.println("\n[formulateBelief,40] belief chosen stochastically: " + belief);
		}
		else{
			belief = makePrediction(t);
			//System.out.println("\n[formulateBelief,44] make prediction: " + belief);
		}
		//System.out.println("\n[formulateBelief,46] return belief: " + belief);
		return belief;
	}
	
	
	/**
	 * Creates prediction for time t
	 * @return prediction
	 */
	public String makePrediction(int time){
		Enumeration<String> e = observations.keys();
		predictions.clear();
		String obs = null;
		maxValue = 0.0;
		double expValue = 0.0;
		//System.out.println("\n[makePrediction,60] entered");
		
		// (1) Add observation to prediction set
		while(e.hasMoreElements()){
			obs = e.nextElement();
			//System.out.println("\n[makePrediction,65] choose existing observation: " + obs);
			
			expValue = MathOperations.roundTo4( (double) observations.get(obs) / time) ;
			//System.out.println("\n[makePrediction,68] expValue of obs: " + expValue);
			
			if( expValue > maxValue){
				maxValue = expValue;
				//System.out.println("\n[makePrediction,72] new maxValue: " + maxValue);
				
				predictions.clear();
				predictions.add(obs);
				//System.out.println("\n[makePrediction,76] clear prediction set and add obs:  " + obs);
			}
			
			else if( expValue  >= ( maxValue - .001) && expValue <= (maxValue + .001) ) {
				predictions.add(obs);
				//System.out.println("\n[makePrediction,81] No new maxValue, only add obs to prediction set: " + obs);
			}
		}
		
		// (2) Choose one out of the prediction set with max value randomly
		if (predictions.size() == 1) {
			belief = predictions.get(0);
			//System.out.println("\n[makePrediction,88] Only one prediction in set. Choose: " + belief);
			return belief;
		
		} else {
			belief = choosePrediction(predictions);
			//System.out.println("\n[makePrediction,93] Selection out of many. Choose: " + belief);
			return belief;
		}
	}
	
	
	/**
	 * Selects prediction out of an array of possible events
	 * 
	 * @param prediction set
	 * @return prediction
	 */
	public String choosePrediction(ArrayList<String> predictions){
		random =MathOperations.roundTo2(Math.random());
		threshold = 0.0;
		//System.out.println("\n[choosePrediction,107] entered. Size of prediction set: " + predictions.size()
		//								+ "\n Random number: " + random);
		
		for(int i = 0; i < predictions.size(); i++){
			threshold += MathOperations.roundTo2( (double) 1 / predictions.size());
			if(threshold > .99){
				threshold = 1.0;
			}
			//System.out.println("\n[choosePrediction,111] New threshold: "+ threshold);
			
			if (random <= threshold){
				//System.out.println("\n[choosePrediction,118] random is below threshold. New prediction: "+ predictions.get(i));
				return predictions.get(i);
			}
		}
		//System.out.println("[ERROR-Beliefs.java] No prediction chosen (choosePrediction) ");
		System.exit(0);
		return null;
	}
	
	/**
	 * Random selection
	 * 
	 * @return
	 */
	public String chooseBeliefStochastically(){
		random =MathOperations.roundTo2(Math.random());
		threshold = 0.0;
		double x = 0.0;
		//System.out.println("\n[chooseBeliefStochastically,136] entered. Length of action set: " + actions.length +
		//								"\nRandom number: " + random);
		
		for(int i = 0; i < actions.length; i++){
			threshold +=   ((double) 1/actions.length);
			if(threshold > .99){
				threshold = 1.0;
			}
			//System.out.println("\n[chooseBeliefStochastically,140] Threshold of action " + i +": " + threshold);
			
			if (random <= threshold){
				//System.out.println("\n[chooseBeliefStochastically,146] Random number is below threshold. Return:  " + actions[i]);
				return String.valueOf(actions[i]);
			}
		}
		// Error - no action chosen
		System.exit(0);
		return null;
	}
	
	/**
	 * Observe environment
	 * 
	 * @param observation
	 */
	public void observe(String observation)
	{
		//System.out.println("\n[observe,161] observation " + observation);
		
		if (!checkState(observation))
		{
			//System.out.println("\n[observe,165] new observation - create new state");
			createState(observation);
		} else{
			//System.out.println("\n[observe,168] known observation - increase counter");
			// set counter +1
			observations.put(		observation, 
												observations.get(observation) + 1 );
			//System.out.println("\n[observe,172] New counter value: " + observations.get(observation));
		}
	}
	
	
	/* (non-Javadoc)
	 * @see knowledgebase.Knowledgebase#checkState(java.lang.String)
	 */
	public boolean checkState(String observation)
	{
		if(observations.containsKey(observation))
			return true;
		else 
			return false;
	}
	
	/* (non-Javadoc)
	 * @see knowledgebase.Knowledgebase#createState(java.lang.String)
	 */
	public void createState(String observation)
	{
		state_counter++;
		//	System.out.println("\n[createState,193] entered. New state counter: " + state_counter
		//								+ "\n new observation added.");
		
		observations.put(observation, 1);
	}
}
