/**
* File: State.java
* --------------------------------------------------------
* Creates a state object for the knowledge base
* (environmental state as perceived and action set / strategies). 
* 
* Author:  Iris Lorscheid
* Date: May, 1st 2014 
*
*/
package knowledgebase;

import java.util.*;

import library.*;

public class State {

	/** State identifier [perc1|perc2|perc3]*/
	private String id; 
	
	/** Hashtable with strategies */
	private Hashtable<String,Strategy> str;
	private double initial_strength;
	private double[] actions;
	
	/**
	 * Constructor
	 * 
	 * @param id 				- unique identifier
	 * @param actions			- possible actions
	 * @param initial_strength	- initial strength of actions
	 */
	public State(	String id, 
					double[] actions,
					double initial_strength
				)
	{
		this.initial_strength = initial_strength;
		this.setActions(actions);
		this.setId(id);
		str = new Hashtable<String,Strategy>();
		for (int i=0; i<actions.length; i++)
		{
			str.put(
					String.valueOf(actions[i]),
					new Strategy(	new String(id + ";" + String.valueOf(actions[i])),
									actions[i],
									initial_strength
								)
					);
		}
	}
	
	/**
	 * Select truthful action for current state
	 * 
	 * @return action
	 */
	public double selectActionTruthfully()
	{
		Enumeration<Strategy> e0 = str.elements();
		Strategy tmp = null;
		while( e0.hasMoreElements())
		{
		 tmp = e0.nextElement();
		 if(tmp.getAction()== 0.0 )
		 {
		 System.out.println("\n [EXCEPTION] Truthful action chosen: " + tmp.getAction());
		 return tmp.getAction();	
		 }
		}
		return 0.0;
	}
	
	/**
	 * Selects an action from strategy set with a probability for every strategy
	 * based on its strength.
	 * 
	 * @return action from chosen strategy
	 */
	public double selectActionStochastically()
	{
		// (1) calculate SUM of strength values
		double strength_sum = 0.00;		
		Enumeration<Strategy> e1 = str.elements();
		while( e1.hasMoreElements())
		{
		 strength_sum= MathOperations.roundTo4(strength_sum + e1.nextElement().getStrength());
		}
		
		// strength sum zero => uniform probability distribution
		
		/*if ( strength_sum == 0){
			for (int i=0; i < actions.length; i++){
				str.get(String.valueOf(actions[i])).setStrength(initial_strength);
				strength_sum+=math_bib.roundTo2(e1.nextElement().getStrength());
			}
		}*/
		
		if ( strength_sum == 0){
			Enumeration<Strategy> e0 = str.elements();
			while( e0.hasMoreElements())
			{
			 e0.nextElement().setStrength(initial_strength);
			 strength_sum= MathOperations.roundTo4(strength_sum + initial_strength);
			}
		}
		
		//System.out.println("Strength SUM ++++ " + strength_sum);
		
		// (2) choose random number [0,1]
		double random = MathOperations.roundTo4(Math.random());

		// (3) choose strategy stochastically
		Enumeration<Strategy> e2 = str.elements();
		double threshold = 0.0;
		Strategy tmp = null;
		while( e2.hasMoreElements())
		{	
			tmp = e2.nextElement();
			threshold = MathOperations.roundTo4(threshold + tmp.calcSelectionProbability(strength_sum));
			
			if(threshold>=0.99){
				threshold = 1;
			}
			
				/*System.out.println("Stragety: " + tmp.getId() + " -- Action: " + tmp.getAction() +
					"-- Strength: " + tmp.getStrength() + "-- threshold: " + threshold
					+ "-- RANDOM NR: " + random);*/
			
			if(random <= threshold)
			{
			 // (4) return action of strategy
			//System.out.println("State.java - line 122 - Strength of chosen action: " + tmp.getStrength());
			 return tmp.getAction();	
			} 
		}
		
		// Error[1]
		System.out.println("ERROR[[1]State.selectActionStochastical()] NO ACTION CHOSEN!");
		System.exit(0);
		return -1;
	}

	/** GETTER & SETTER **/
	
	public Hashtable<String, Strategy> getStr() {
		return str;
	}
	public void setStr(Hashtable<String, Strategy> str) {
		this.str = str;
	}
	public void setId(String id) {
		this.id = id;
	}
	public String getId() {
		return id;
	}

	public double[] getActions() {
		return actions;
	}

	public void setActions(double[] actions) {
		this.actions = actions;
	}
}
