/**
* File: FicticiousPlay.java
* --------------------------------------------------------
* Implements the ficticious play learning model. 
* 
* Author:  Iris Lorscheid
* Date: May, 1st 2014 
*
*/

package learning;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;

import library.ArrayOperations;
import library.MathOperations;
import library.ProfitCalculation;
import library.ResourceAllocation;
import mechanism.GrovesMechanism;
import knowledgebase.Beliefs;
import observer.Observer;

public class FicticiousPlay extends LearningAlgorithm {

	private Observer observer;
	private Hashtable<String, ArrayList> bestResponses;
	private ArrayList<Double> actions;
	private double pMin, pMax, incrRepP;
	private double response;
	
	/**
	 * @throws IOException
	 */
	public FicticiousPlay() throws IOException{
		super();
	}
	
	/**
	 * Initializes ficticious play algortihm for given setting
	 * 
	 * @param pMin
	 * @param prodScaleRep
	 * @param incrRepP
	 * @throws IOException 
	 */
	public void initializeFicticiousPlay 	(	double pMin, int prodScaleRep, 
																		double incrRepP, Observer observer) throws IOException {
		this.pMin = pMin;
		this.pMax = MathOperations.roundTo1( pMin + ((prodScaleRep-1) * .1));
		this.incrRepP = incrRepP;
		bestResponses = new Hashtable<String,ArrayList>();
		actions = new ArrayList<Double>();
		response = 0.0;
		this.observer = observer;
	}

	/**
	 * @param prediction
	 * @return
	 * @throws IOException 
	 */
	public double respond (double realP, double prediction) {
		actions.clear();
		double report =0.0;
		double expPayment = 0.0;
		double expProfit = 0.0;
		int expResources = 0;
		double expProfitOpponent = 0.0;
		int expResourcesOpponent = 0;
		double maxValue = 0.0;
		//System.out.println("\n[respond,56] entered. realP: " + realP + ", prediction: " + prediction);
		
		for (report = pMin; report <= pMax; report=MathOperations.roundTo1(report+incrRepP)){
			//System.out.println("\n[respond,59] report_i:  " + report);
		
			expResources = ResourceAllocation.allocate(report, prediction, observer.getResources(),observer);
			//System.out.println("\n[respond,62] expected resources: " + expResources);
			
			expProfit = ProfitCalculation.calcProfit(expResources, realP, observer);
			//System.out.println("\n[respond,65] expected profit: " + expProfit);
			
			expResourcesOpponent = ResourceAllocation.allocate(prediction, report, observer.getResources(),observer);
			
			expProfitOpponent = ProfitCalculation.calcProfit(expResourcesOpponent, prediction, observer);
			
			try {
				if(report == realP)
						expPayment = GrovesMechanism.calcCompensation(
														expProfit, 
														expProfitOpponent, 
														observer.getDiscount(),
														true);
				else
						expPayment = GrovesMechanism.calcCompensation(
														expProfit, 
														expProfitOpponent, 
														observer.getDiscount(),
														false);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			
			if ( 	expPayment >= ( maxValue - .001) 
					&& 
					expPayment<= (maxValue + .001) 
					){
				actions.add(report);
				//System.out.println("\n[respond,73] new action with same maxValue added: " + report);
			}
			
			if (expPayment > maxValue){
				maxValue = expPayment;
				//System.out.println("\n[respond,77] new maxValue:  " + maxValue);
				
				actions.clear();
				actions.add(report);
				//System.out.println("\n[respond,81] actions cleared, add new report: " + report);
			}
		}
		
		double random = Math.random();
		//System.out.println("\n[respond,86] random number: " + random);
		
		double threshold = 0.0;
		for(int i = 0; i < actions.size(); i++){
			//System.out.println("\n[respond,90] action size: " + actions.size());
			
			threshold += (double) 1 / actions.size();
			if (threshold > .99)
				threshold = 1;
			//System.out.println("\n[respond,95] threshold: " + threshold);
			
			if (random < threshold){
				response = (double) actions.get(i);
				//System.out.println("\n[respond,99] random < threshold -- response: " + response);
				return response;
			}
		}
		
	 return response;
	}

	/**
	 * @param observedReportOpponent
	 * @param beliefs
	 */
	public void updateKB (double observedReportOpponent, Beliefs beliefs) {
		//System.out.println("\n[updateKB,102_fictPlay] observed report: " + observedReportOpponent);
		beliefs.observe(String.valueOf(observedReportOpponent));
	}
	
}
