/**
* File: ObserverLM2.java
* --------------------------------------------------------
* Superclass for the Observer class of LM2 (scheduling and reporting).
* 
* Author:  Iris Lorscheid
* Date: May, 1st 2014 
*
*/

package observer;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

import knowledgebase.Knowledgebase;
import library.ArrayOperations;
import library.MathOperations;
import library.ProfitCalculation;
import library.ResourceAllocation;
import mechanism.GrovesMechanism;
import agents.DivisionManagerLM2;
import agents.Headquarter;

public class ObserverLM2 extends Observer {
	private ArrayList<DivisionManagerLM2> divManAgents;
	private double lamda;
	 
	/**
	 * Constructor
	 * 
	 * @param identifier
	 * @param resources
	 * @param pMin
	 * @param prodScale
	 * @param opponentProdScale
	 * @param incrRealP
	 * @param opponentIncrRealP
	 * @param prodScaleRep
	 * @param opponentProdScaleRep
	 * @param incrRepP
	 * @param opponentIncrRepP
	 * @param optCapacity
	 * @param maxCapacity
	 * @param reductProd
	 * @param discount
	 * @param numberOfTicks
	 * @param numberOfRuns
	 * @param verificationOutput
	 * @param designPoint
	 * @param lamda
	 * @throws IOException
	 */
	public ObserverLM2(String identifier, int resources, double pMin, int prodScale, int opponentProdScale,
										double incrRealP, double opponentIncrRealP, int prodScaleRep, int opponentProdScaleRep,
										double incrRepP, double opponentIncrRepP, 	double optCapacity,	double maxCapacity,
										double reductProd, double discount, int numberOfTicks, int numberOfRuns, 
										FileWriter verificationOutput, int designPoint, double lamda
										) throws IOException{
		
		super(			resources, pMin, prodScale, opponentProdScale,
							incrRealP, opponentIncrRealP, prodScaleRep, opponentProdScaleRep,
							incrRepP,  opponentIncrRepP,  optCapacity, maxCapacity,
							reductProd, discount, numberOfTicks, numberOfRuns, 
							verificationOutput, designPoint);
		
		this.identifier = identifier;
		this.divManAgents = new ArrayList<DivisionManagerLM2>();		
		this.lamda = lamda;
		
		checkSetting();
		createAgents();
		initializeLA();
		runSimulationExperiment();
	}
	
	/**
	 * 
	 */
	public void initializeLA(){
		divManAgents.get(0).getLa().setLambda(lamda);
		divManAgents.get(1).getLa().setLambda(lamda);
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#resetRun()
	 */
	public void resetRun(){
		divManAgents.get(0).setKb(null);
		divManAgents.get(0).setKb(
					new Knowledgebase(		ArrayOperations.generateArray(prodScaleRep, pMin, incrRepP),
																	MathOperations.roundTo4( (double) 1 / (double) prodScaleRep),
																	new String("divMan-A-" + identifier)
				));
		
		divManAgents.get(1).setKb(null);
		divManAgents.get(1).setKb(
					new Knowledgebase(		ArrayOperations.generateArray(prodScaleRep, pMin, incrRepP),
																	MathOperations.roundTo4( (double) 1 / (double) prodScaleRep),
																	new String("divMan-B-" + identifier)
				));
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#createHeadquarter()
	 */
	public void createHeadquarter(){
		hqAgent = new Headquarter(
				new String("hqAgent-X" + identifier),
				this,
				resources,
				absOptCapacity,
				absMaxCapacity,
				reductProd,
				discount,
				verificationOutput
				);
	}

	/* (non-Javadoc)
	 * @see observer.Observer#createDivManagers()
	 */
	public void createDivManagers() throws IOException{
		String id = "A";												// to make it easier to distinguish between setting ID and agent ID
		String identifier_tmp = new String("divMan-" + id + "-" + identifier);
			
		divManAgents.add(new DivisionManagerLM2(			// the same parameters for both agents, accord. to lab.exp.
															identifier_tmp,
															reductProd, 
															absOptCapacity, 
															absMaxCapacity, 
															pMin, 
															prodScaleRep, 
															incrRepP, 
															verificationOutput,
															this
															)
												);
		
		id = "B";
		identifier_tmp = new String("divMan-" + id + "-" + identifier);
		
		divManAgents.add(new DivisionManagerLM2(			// the same parameters for both agents, accord. to lab.exp.
															identifier_tmp,
															reductProd, 
															absOptCapacity, 
															absMaxCapacity, 
															pMin, 
															opponentProdScaleRep, 
															opponentIncrRepP, 
															verificationOutput,
															this
															)
												);
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#round()
	 */
	public void round() throws IOException {
		
		// Division managers receive real productivity and report
		double report1 = divManAgents.get(0).reportProductivity(realProductivity());
		double report2 = divManAgents.get(1).reportProductivity(realProductivityOpponent());
		
		// Headquarter allocates resources to divisions
		// Division managers produce
		// Headquarter calculates profit achieved.
		divManAgents.get(0).produce(hqAgent.allocResources(report1, report2));
		divManAgents.get(1).produce(hqAgent.allocResources(report2, report1));

		// Headquarter pays compensation to divisions, calculation by Groves
		if(report1 == divManAgents.get(0).getProdValue()){
			divManAgents.get(0).receiveCompensation(
						hqAgent.payCompensation(	divManAgents.get(0).getProfit(),
																				ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																				report2, 
																																				report1,
																																				resources,
																																				this),
																																	report2,
																																	this
																																	),												
																				discount, true) );
			} else{
				divManAgents.get(0).receiveCompensation(
						hqAgent.payCompensation(	divManAgents.get(0).getProfit(),
																				ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																				report2, 
																																				report1,
																																				resources,
																																				this),
																																	report2,
																																	this
																																	),												
																				discount, false) );
			}
		
		
			if(report2 == divManAgents.get(1).getProdValue()){
			divManAgents.get(1).receiveCompensation(
						hqAgent.payCompensation(	divManAgents.get(1).getProfit(),
																				ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																				report1, 
																																				report2,
																																				resources,
																																				this),
																																		report1,
																																		this
																																		),												
																				discount, 
																				true ) );
			
		} else{
			divManAgents.get(1).receiveCompensation(
					hqAgent.payCompensation(	divManAgents.get(1).getProfit(),
																			ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																			report1, 
																																			report2,
																																			resources,
																																			this),
																																	report1,
																																	this
																																	),												
																			discount, 
																			false ) );
			}
			}

	/* (non-Javadoc)
	 * @see observer.Observer#observeSuccess()
	 */
	public void observeSuccess() throws IOException{
				// profit calculation
				profit = hqAgent.calcCompanyProfit(
																	divManAgents.get(0).getProfit(), 
																	divManAgents.get(1).getProfit());
		
				// company loss by manipulated reports
				loss = calcCompanyLoss();	
						
				// relative compensation		
				relativeCompensation();
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#relativeCompensation()
	 */
	public void relativeCompensation() throws IOException{
		 double maxCompensation1 = GrovesMechanism.calcCompensation(
					
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(0).getProdValue(), 
							divManAgents.get(1).getRepProd(),
							resources,
							this),
							divManAgents.get(0).getProdValue(),
							this
							),												
																			
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(1).getRepProd(), 
							divManAgents.get(0).getProdValue(),
							resources,
							this),
							divManAgents.get(1).getRepProd(),
							this
							),				
							
						discount,
						true);
		
		
		
		 double minPayment1 = calcMinPayment(	divManAgents.get(0).getProdValue(), 
					divManAgents.get(1).getRepProd() 		
					);
		 double relComp1_1 = relCompensation1;
					
		 /*relCompensation1 = MathOperations.roundTo2(
				 divManAgents.get(0).getCompensation()  / maxCompensation1 
				 );*/
		 relCompensation1 = MathOperations.roundTo2( 
				 																	(	  divManAgents.get(0).getCompensation() 
				 																			- 
				 																			minPayment1
				 																	)
				 																	/
				 																	(	  maxCompensation1 
				 																			- 
				 																			minPayment1 
				 																	) 
				 															);
				 															
		//double d1 = relCompensation1 - relComp1_1;
		 if(relCompensation1 <0)
			 relCompensation1 = 0;
		 
		 if (relCompensation1 !=0)
			divManAgents.get(0).updateKB(relCompensation1);
		 
		

		
		 double maxCompensation2 = GrovesMechanism.calcCompensation(
					
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(1).getProdValue(), 
							divManAgents.get(0).getRepProd(),
							resources,
							this),
							divManAgents.get(1).getProdValue(),
							this
							),												
																			
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(0).getRepProd(), 
							divManAgents.get(1).getProdValue(),
							resources,
							this),
							divManAgents.get(0).getRepProd(),
							this
							),				
							
						discount,
						true);
		 
		 double relComp2_1 = relCompensation2;
		 
		 double minPayment2 = calcMinPaymentOpp(	divManAgents.get(1).getProdValue(), 
				 																				divManAgents.get(0).getRepProd() 		
																					);
					
		 /* relCompensation2 = MathOperations.roundTo2(
				 divManAgents.get(1).getCompensation()  / maxCompensation2 
				 );*/
		 relCompensation2 = MathOperations.roundTo2( 
				 																	(	  divManAgents.get(1).getCompensation() 
				 																			- 
				 																			minPayment2
				 																	)
				 																	/
				 																	(	  maxCompensation2 
				 																			- 
				 																			minPayment2 
				 																	) 
				 															);
				 															
		 if(relCompensation2 <0)
			 relCompensation2 = 0; 
		// double d2 = relCompensation2 - relComp2_1;
		 if (relCompensation2!=0)
					divManAgents.get(1).updateKB(relCompensation2);

	}

	/* (non-Javadoc)
	 * @see observer.Observer#observeReportingBehavior()
	 */
	public void observeReportingBehavior(){
			delta1= MathOperations.roundTo2(divManAgents.get(0).getRepProd() - divManAgents.get(0).getProdValue());
			relDelta1 = MathOperations.roundTo2(calcRelDelta(0));
			delta2= MathOperations.roundTo2(divManAgents.get(1).getRepProd() - divManAgents.get(1).getProdValue());
			relDelta2 = MathOperations.roundTo2(calcRelDelta(1));
	}
	

	/* (non-Javadoc)
	 * @see observer.Observer#calcTruthfulProfit()
	 */
	public double calcTruthfulProfit() throws IOException{
		
		return 
		ProfitCalculation.calcProfit(ResourceAllocation.allocate(	
																										divManAgents.get(0).getProdValue(), 
																										divManAgents.get(1).getProdValue(),
																										resources,
																										this),
															divManAgents.get(0).getProdValue(),
															this)
		
		+
		
		ProfitCalculation.calcProfit(ResourceAllocation.allocate(	
																										divManAgents.get(1).getProdValue(), 
																										divManAgents.get(0).getProdValue(),
																										resources,
																										this),
																divManAgents.get(1).getProdValue(),
																this);
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#calcRelDelta(int)
	 */
	public double calcRelDelta(int index){
		if(divManAgents.get(index).getRepProd() < divManAgents.get(index).getProdValue())
			
			return			(divManAgents.get(index).getRepProd()  - divManAgents.get(index).getProdValue())
								/ (divManAgents.get(index).getProdValue() - pMin );
		
		else if(divManAgents.get(index).getRepProd() == divManAgents.get(index).getProdValue())
			
			return  0.0;
		
		else if(divManAgents.get(index).getRepProd() > divManAgents.get(index).getProdValue())
			
			return 			(divManAgents.get(index).getRepProd()  - divManAgents.get(index).getProdValue())
								/ (pMax - divManAgents.get(index).getProdValue() );
		
		return -1111.1;
	}

	/* (non-Javadoc)
	 * @see observer.Observer#printOutput()
	 */
	public void printOutput() throws IOException{
	verificationOutput.append(	 
			//"\n" + designPoint +
			";" + divManAgents.get(0).getProdValue() + 
			";" + divManAgents.get(0).getRepProd() + 
			";" + delta1 +
			";" + relDelta1+
			";" + divManAgents.get(0).getResourcesAllocated() +
			";" +divManAgents.get(0).getProfit()+
			";" + divManAgents.get(0).getCompensation() +
			";" + relCompensation1+
			
			";" + divManAgents.get(1).getProdValue() +
			";" + divManAgents.get(1).getRepProd() +
			";" + delta2+
			";" + relDelta2 +
			";" + divManAgents.get(1).getResourcesAllocated() +
			";" + divManAgents.get(1).getProfit()+
			";" + divManAgents.get(1).getCompensation() +
			";" + relCompensation2 +
			
			";" + profit+
			";" + calcTruthfulProfit() +
			";" + loss +
			";" + hits +			
			
			";" + resources +
			";" + pMin +
			";" + prodScale +
			";" + optCapacity+
			";" + maxCapacity +
			";" + reductProd +
			";" + discount+
			/*";" +
			ProfitCalculation.calcProfit(ResourceAllocation.allocate(
					divManAgents.get(1).getProdValue(), 
					divManAgents.get(0).getRepProd(),
					resources,
					this),
					divManAgents.get(1).getProdValue(),
					this
					)+ */
			";" + lamda
			+	";"  //k
			+ ";2" + // 2 for LM2
			";" + numberOfRuns +
			";" + numberOfTicks
			);
	}
}
