/**
* File: ObserverLM3.java
* --------------------------------------------------------
* Superclass for the Observer class of LM3 (scheduling and reporting).
* 
* Author:  Iris Lorscheid
* Date: May, 1st 2014 
*
*/

package observer;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

import library.MathOperations;
import library.ProfitCalculation;
import library.ResourceAllocation;
import mechanism.GrovesMechanism;
import agents.DivisionManagerLM3;
import agents.Headquarter;

public class ObserverLM3 extends Observer {
	private ArrayList<DivisionManagerLM3> divManAgents;
	private double lamda;
	 
	/**
	 * Constructor
	 * 
	 * @param identifier
	 * @param resources
	 * @param pMin
	 * @param prodScale
	 * @param opponentProdScale
	 * @param incrRealP
	 * @param opponentIncrRealP
	 * @param prodScaleRep
	 * @param opponentProdScaleRep
	 * @param incrRepP
	 * @param opponentIncrRepP
	 * @param optCapacity
	 * @param maxCapacity
	 * @param reductProd
	 * @param discount
	 * @param numberOfTicks
	 * @param numberOfRuns
	 * @param verificationOutput
	 * @param designPoint
	 * @throws IOException
	 */
	public ObserverLM3(String identifier, int resources, double pMin, int prodScale, int opponentProdScale,
										double incrRealP, double opponentIncrRealP, int prodScaleRep, int opponentProdScaleRep,
										double incrRepP, double opponentIncrRepP, 	double optCapacity,	double maxCapacity,
										double reductProd, double discount, int numberOfTicks, int numberOfRuns, 
										FileWriter verificationOutput, int designPoint, double lamda
										) throws IOException{
		
		super(			resources, pMin, prodScale, opponentProdScale,
							incrRealP, opponentIncrRealP, prodScaleRep, opponentProdScaleRep,
							incrRepP,  opponentIncrRepP,  optCapacity, maxCapacity,
							reductProd, discount, numberOfTicks, numberOfRuns, 
							verificationOutput, designPoint);
		
		this.identifier = identifier;
		this.divManAgents = new ArrayList<DivisionManagerLM3>();
		this.lamda = lamda;
		checkSetting();
		createAgents();
		initializeLA();
		runSimulationExperiment();
	}
	
	/**
	 * @throws IOException 
	 * 
	 */
	public void initializeLA() throws IOException{
		divManAgents.get(0).getLa().initializeFicticiousPlay(	pMin, prodScaleRep, 
																										incrRepP, this);
		divManAgents.get(1).getLa().initializeFicticiousPlay( 	pMin, opponentProdScaleRep, 
																										opponentIncrRepP, this);
	}
	
	
	/* (non-Javadoc)
	 * @see observer.Observer#createHeadquarter()
	 */
	public void createHeadquarter(){
		hqAgent = new Headquarter(
				new String("hqAgent-X" + identifier),
				this,
				resources,
				absOptCapacity,
				absMaxCapacity,
				reductProd,
				discount,
				verificationOutput
				);
	}

	/* (non-Javadoc)
	 * @see observer.Observer#createDivManagers()
	 */
	public void createDivManagers() throws IOException{
		String id = "A";												// to make it easier to distinguish between setting ID and agent ID
		String identifier_tmp = new String("divMan-" + id + "-" + identifier);
			
		divManAgents.add(new DivisionManagerLM3(			// the same parameters for both agents, accord. to lab.exp.
															identifier_tmp,
															reductProd, 
															absOptCapacity, 
															absMaxCapacity, 
															pMin, 
															prodScaleRep, 
															incrRepP, 
															verificationOutput,
															this
															)
												);
		
		id = "B";
		identifier_tmp = new String("divMan-" + id + "-" + identifier);
		
		divManAgents.add(new DivisionManagerLM3(			// the same parameters for both agents, accord. to lab.exp.
															identifier_tmp,
															reductProd, 
															absOptCapacity, 
															absMaxCapacity, 
															pMin, 
															opponentProdScaleRep, 
															opponentIncrRepP, 
															verificationOutput,
															this
															)
												);
	}
	
	/**
	 * Schedules one simulation run
	 * ------------------------------------------------
	 * @throws IOException
	 */
	public void runSimulationExperiment() throws IOException{
		for(int index2 = 0; index2 < numberOfRuns; index2++){
		System.out.println("\n Observer " + identifier + " run no. " + index2);
		
		for (int index1 = 0; index1 < numberOfTicks; index1++){
			
			verificationOutput.append("\n" + index1 + ";" + index2);
			
			// 1 simulation round (tick)
			round(index1);
			
			// observes reporting behavior of division manager agents
			observeReportingBehavior();
			
			// observes equilibrium truthful reporting 
			if(delta1 < 0.001 && delta1 > -0.001 && delta2 < 0.001 && delta2 >-0.001)
				hits++;
			
			// observes division and company success (profit and losses)
			observeSuccess();
			
			// write observations in output file
			printOutput();
			
			relCompensation1 = 0.0;	 
			relCompensation2 = 0.0;
		}
		hits = 0;
		}
		//matrix.close();
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#round()
	 */
	public void round(int tick) throws IOException {
		//System.out.println("\n[round,168] round in tick: " + tick);

		// Division managers receive real productivity and report
		double report1 = divManAgents.get(0).reportProductivity(	realProductivity(), 
																													Double.valueOf(divManAgents.get(0).getBeliefs().formulateBelief(tick))
																												);
		//System.out.println("\n[round,174] report 1: " + report1);

		double report2 = divManAgents.get(1).reportProductivity(	realProductivityOpponent(), 
																													Double.valueOf(divManAgents.get(1).getBeliefs().formulateBelief(tick))		
																												);
		//System.out.println("\n[round,179] report 2: " + report2);
		
		// Headquarter allocates resources to divisions
		// Division managers produce
		// Headquarter calculates profit achieved.
		divManAgents.get(0).produce(hqAgent.allocResources(report1, report2));
		divManAgents.get(1).produce(hqAgent.allocResources(report2, report1));

		// Headquarter pays compensation to divisions, calculation by Groves
		if(report1 == divManAgents.get(0).getProdValue()){
			divManAgents.get(0).receiveCompensation(
						hqAgent.payCompensation(	divManAgents.get(0).getProfit(),
																				ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																				report2, 
																																				report1,
																																				resources,
																																				this),
																																	report2,
																																	this
																																	),												
																				discount, true) );
			} else{
				divManAgents.get(0).receiveCompensation(
						hqAgent.payCompensation(	divManAgents.get(0).getProfit(),
																				ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																				report2, 
																																				report1,
																																				resources,
																																				this),
																																	report2,
																																	this
																																	),												
																				discount, false) );
			}
		
		
			if(report2 == divManAgents.get(1).getProdValue()){
			divManAgents.get(1).receiveCompensation(
						hqAgent.payCompensation(	divManAgents.get(1).getProfit(),
																				ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																				report1, 
																																				report2,
																																				resources,
																																				this),
																																		report1,
																																		this
																																		),												
																				discount, 
																				true ) );
			
		} else{
			divManAgents.get(1).receiveCompensation(
					hqAgent.payCompensation(	divManAgents.get(1).getProfit(),
																			ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																			report1, 
																																			report2,
																																			resources,
																																			this),
																																	report1,
																																	this
																																	),												
																			discount, 
																			false ) );
			}

		//System.out.println("\n[round,212] DivManager 0 update KB.");
		divManAgents.get(0).updateKB(report2);
		
		//System.out.println("\n[round,215] DivManager 1 update KB.");
		divManAgents.get(1).updateKB(report1);	
	}

	/* (non-Javadoc)
	 * @see observer.Observer#observeSuccess()
	 */
	public void observeSuccess() throws IOException{
				// profit calculation
				profit = hqAgent.calcCompanyProfit(
																	divManAgents.get(0).getProfit(), 
																	divManAgents.get(1).getProfit());
		
				// company loss by manipulated reports
				loss = calcCompanyLoss();	
						
				// relative compensation		
				relativeCompensation();
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#relativeCompensation()
	 */
	public void relativeCompensation() throws IOException{
		 double maxCompensation1 = GrovesMechanism.calcCompensation(
					
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(0).getProdValue(), 
							divManAgents.get(1).getRepProd(),
							resources,
							this),
							divManAgents.get(0).getProdValue(),
							this
							),												
																			
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(1).getRepProd(), 
							divManAgents.get(0).getProdValue(),
							resources,
							this),
							divManAgents.get(1).getRepProd(),
							this
							),				
							
						discount,
						true);
				
		 double minPayment1 = calcMinPayment(	divManAgents.get(0).getProdValue(), 
					divManAgents.get(1).getRepProd() 		
					);
		 double relComp1_1 = relCompensation1;
					
		 relCompensation1 = MathOperations.roundTo2(
				 divManAgents.get(0).getCompensation()  / maxCompensation1 
				 );
		/* relCompensation1 = MathOperations.roundTo2( 
				 																	(	  divManAgents.get(0).getCompensation() 
				 																			- 
				 																			minPayment1
				 																	)
				 																	/
				 																	(	  maxCompensation1 
				 																			- 
				 																			minPayment1 
				 																	) 
				 															);
				 															*/

		 //double d1 = relCompensation1 - relComp1_1;

		//if (relCompensation1 !=0)
		 //divManAgents.get(0).updateKB(relCompensation1);
		// divManAgents.get(0).updateKB(divManAgents.get(0).getCompensation());
			
		 
		 double maxCompensation2 = GrovesMechanism.calcCompensation(
					
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(1).getProdValue(), 
							divManAgents.get(0).getRepProd(),
							resources,
							this),
							divManAgents.get(1).getProdValue(),
							this
							),												
																			
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(0).getRepProd(), 
							divManAgents.get(1).getProdValue(),
							resources,
							this),
							divManAgents.get(0).getRepProd(),
							this
							),				
							
						discount,
						true);
		 
		 double relComp2_1 = relCompensation2;
		 
		 double minPayment2 = calcMinPaymentOpp(	divManAgents.get(1).getProdValue(), 
				 																				divManAgents.get(0).getRepProd() 		
																					);
					
		 relCompensation2 = MathOperations.roundTo2(
				 divManAgents.get(1).getCompensation()  / maxCompensation2 
				 );
		/* relCompensation2 = MathOperations.roundTo2( 
				 																	(	  divManAgents.get(1).getCompensation() 
				 																			- 
				 																			minPayment2
				 																	)
				 																	/
				 																	(	  maxCompensation2 
				 																			- 
				 																			minPayment2 
				 																	) 
				 															);
				 															*/
		
		//double d2 = relCompensation2 - relComp2_1;
		 //if (relCompensation2 !=0){
		// divManAgents.get(1).updateKB(relCompensation2);
		// divManAgents.get(1).updateKB(divManAgents.get(1).getCompensation());
		 //}
	}

	/* (non-Javadoc)
	 * @see observer.Observer#observeReportingBehavior()
	 */
	public void observeReportingBehavior(){
			delta1= MathOperations.roundTo2(divManAgents.get(0).getRepProd() - divManAgents.get(0).getProdValue());
			relDelta1 = MathOperations.roundTo2(calcRelDelta(0));
			delta2= MathOperations.roundTo2(divManAgents.get(1).getRepProd() - divManAgents.get(1).getProdValue());
			relDelta2 = MathOperations.roundTo2(calcRelDelta(1));
	}
	

	/* (non-Javadoc)
	 * @see observer.Observer#calcTruthfulProfit()
	 */
	public double calcTruthfulProfit() throws IOException{
		
		return 
		ProfitCalculation.calcProfit(ResourceAllocation.allocate(	
																										divManAgents.get(0).getProdValue(), 
																										divManAgents.get(1).getProdValue(),
																										resources,
																										this),
															divManAgents.get(0).getProdValue(),
															this)
		
		+
		
		ProfitCalculation.calcProfit(ResourceAllocation.allocate(	
																										divManAgents.get(1).getProdValue(), 
																										divManAgents.get(0).getProdValue(),
																										resources,
																										this),
																divManAgents.get(1).getProdValue(),
																this);
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#calcRelDelta(int)
	 */
	public double calcRelDelta(int index){
		if(divManAgents.get(index).getRepProd() < divManAgents.get(index).getProdValue())
			
			return			(divManAgents.get(index).getRepProd()  - divManAgents.get(index).getProdValue())
								/ (divManAgents.get(index).getProdValue() - pMin );
		
		else if(divManAgents.get(index).getRepProd() == divManAgents.get(index).getProdValue())
			
			return  0.0;
		
		else if(divManAgents.get(index).getRepProd() > divManAgents.get(index).getProdValue())
			
			return 			(divManAgents.get(index).getRepProd()  - divManAgents.get(index).getProdValue())
								/ (pMax - divManAgents.get(index).getProdValue() );
		
		return -1111.1;
	}

	/* (non-Javadoc)
	 * @see observer.Observer#printOutput()
	 */
	public void printOutput() throws IOException{
	verificationOutput.append(	 
			//"\n" + designPoint +
			";" + divManAgents.get(0).getProdValue() + 
			";" + divManAgents.get(0).getRepProd() + 
			";" + delta1 +
			";" + relDelta1+
			";" + divManAgents.get(0).getResourcesAllocated() +
			";" +divManAgents.get(0).getProfit()+
			";" + divManAgents.get(0).getCompensation() +
			";" + relCompensation1+
			
			";" + divManAgents.get(1).getProdValue() +
			";" + divManAgents.get(1).getRepProd() +
			";" + delta2+
			";" + relDelta2 +
			";" + divManAgents.get(1).getResourcesAllocated() +
			";" + divManAgents.get(1).getProfit()+
			";" + divManAgents.get(1).getCompensation() +
			";" + relCompensation2 +
			
			";" + profit+
			";" + calcTruthfulProfit() +
			";" + loss +
			";" + hits +			
			
			";" + resources +
			";" + pMin +
			";" + prodScale +
			";" + optCapacity+
			";" + maxCapacity +
			";" + reductProd +
			";" + discount+
			/*";" +
			ProfitCalculation.calcProfit(ResourceAllocation.allocate(
					divManAgents.get(1).getProdValue(), 
					divManAgents.get(0).getRepProd(),
					resources,
					this),
					divManAgents.get(1).getProdValue(),
					this
					)+ */
			";"  // lamda
			+	";"  //k
			+ ";3" + // 3 for LM3
			";" + numberOfRuns +
			";" + numberOfTicks +
			";" + divManAgents.get(0).getBeliefs().getBelief()+
			";" + divManAgents.get(1).getBeliefs().getBelief()+
			";" + divManAgents.get(0).getBeliefs().getMaxValue() +
			";" + divManAgents.get(1).getBeliefs().getMaxValue()
			);
	}
}
