/**
* File: ObserverZI.java
* --------------------------------------------------------
* Superclass for the Observer class of ZI (scheduling and reporting).
* 
* Author:  Iris Lorscheid
* Date: May, 1st 2014 
*
*/

package observer;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

import library.MathOperations;
import library.ProfitCalculation;
import library.ResourceAllocation;
import mechanism.GrovesMechanism;
import agents.*;

public class ObserverZI extends Observer{
	private ArrayList<DivisionManagerZI> divManAgents;
	 
	/**
	 * Constructor
	 * 
	 * @param identifier
	 * @param resources
	 * @param pMin
	 * @param prodScale
	 * @param opponentProdScale
	 * @param incrRealP
	 * @param opponentIncrRealP
	 * @param prodScaleRep
	 * @param opponentProdScaleRep
	 * @param incrRepP
	 * @param opponentIncrRepP
	 * @param optCapacity
	 * @param maxCapacity
	 * @param reductProd
	 * @param discount
	 * @param numberOfTicks
	 * @param numberOfRuns
	 * @param verificationOutput
	 * @param designPoint
	 * @throws IOException
	 */
	public ObserverZI(	String identifier, int resources, double pMin, int prodScale, int opponentProdScale,
										double incrRealP, double opponentIncrRealP, int prodScaleRep, int opponentProdScaleRep,
										double incrRepP, double opponentIncrRepP, 	double optCapacity,	double maxCapacity,
										double reductProd, double discount, int numberOfTicks, int numberOfRuns, 
										FileWriter verificationOutput, int designPoint
										) throws IOException{
		
		super(			resources, pMin, prodScale, opponentProdScale,
							incrRealP, opponentIncrRealP, prodScaleRep, opponentProdScaleRep,
							incrRepP,  opponentIncrRepP,  optCapacity, maxCapacity,
							reductProd, discount, numberOfTicks, numberOfRuns, 
							verificationOutput, designPoint);
		this.identifier = identifier;	
		this.divManAgents = new ArrayList<DivisionManagerZI>();
		
		checkSetting();
		createAgents();
		runSimulationExperiment();
	}
	

	/* (non-Javadoc)
	 * @see observer.Observer#createHeadquarter()
	 */
	public void createHeadquarter(){
		hqAgent = new Headquarter(
				new String("hqAgent-X" + identifier),
				this,
				resources,
				absOptCapacity,
				absMaxCapacity,
				reductProd,
				discount,
				verificationOutput
				);
	}

	/* (non-Javadoc)
	 * @see observer.Observer#createDivManagers()
	 */
	public void createDivManagers(){
		String id = "A";												// to make it easier to distinguish between setting ID and agent ID
		String identifier_tmp = new String("divMan-" + id + "-" + identifier);
			
		divManAgents.add(new DivisionManagerZI(			// the same parameters for both agents, accord. to lab.exp.
															identifier_tmp,
															reductProd, 
															absOptCapacity, 
															absMaxCapacity, 
															pMin, 
															prodScaleRep, 
															incrRepP, 
															verificationOutput,
															this
															)
												);
		
		id = "B";
		identifier_tmp = new String("divMan-" + id + "-" + identifier);
		
		divManAgents.add(new DivisionManagerZI(			// the same parameters for both agents, accord. to lab.exp.
															identifier_tmp,
															reductProd, 
															absOptCapacity, 
															absMaxCapacity, 
															pMin, 
															opponentProdScaleRep, 
															opponentIncrRepP, 
															verificationOutput,
															this
															)
												);
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#round()
	 */
	public void round() throws IOException {
		
		// Division managers receive real productivity and report
		double report1 =  divManAgents.get(0).reportProductivity(realProductivity());
		double report2 = divManAgents.get(1).reportProductivity(realProductivityOpponent());
	
		// Headquarter allocates resources to divisions
		// Division managers produce
		// Headquarter calculates profit achieved.
		divManAgents.get(0).produce(hqAgent.allocResources(report1, report2));
		divManAgents.get(1).produce(hqAgent.allocResources(report2, report1));
			
		// Headquarter pays compensation to divisions, calculation by Groves
		if(report1 == divManAgents.get(0).getProdValue()){
			divManAgents.get(0).receiveCompensation(
						hqAgent.payCompensation(	divManAgents.get(0).getProfit(),
																				ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																				report2, 
																																				report1,
																																				resources,
																																				this),
																																	report2,
																																	this
																																	),												
																				discount, true) );
			} else{
				divManAgents.get(0).receiveCompensation(
						hqAgent.payCompensation(	divManAgents.get(0).getProfit(),
																				ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																				report2, 
																																				report1,
																																				resources,
																																				this),
																																	report2,
																																	this
																																	),												
																				discount, false) );
			}
		
		
			if(report2 == divManAgents.get(1).getProdValue()){
			divManAgents.get(1).receiveCompensation(
						hqAgent.payCompensation(	divManAgents.get(1).getProfit(),
																				ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																				report1, 
																																				report2,
																																				resources,
																																				this),
																																		report1,
																																		this
																																		),												
																				discount, 
																				true ) );
			
		} else{
			divManAgents.get(1).receiveCompensation(
					hqAgent.payCompensation(	divManAgents.get(1).getProfit(),
																			ProfitCalculation.calcProfit(ResourceAllocation.allocate(
																																			report1, 
																																			report2,
																																			resources,
																																			this),
																																	report1,
																																	this
																																	),												
																			discount, 
																			false ) );
			}
	}

	/* (non-Javadoc)
	 * @see observer.Observer#observeSuccess()
	 */
	public void observeSuccess() throws IOException{
				// profit calculation
				profit = hqAgent.calcCompanyProfit(
																	divManAgents.get(0).getProfit(), 
																	divManAgents.get(1).getProfit());
		
				// company loss by manipulated reports
				loss = calcCompanyLoss();	
						
				// relative compensation		
				relativeCompensation();
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#relativeCompensation()
	 */
	public void relativeCompensation() throws IOException{
		 double maxCompensation1 = GrovesMechanism.calcCompensation(
					
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(0).getProdValue(), 
							divManAgents.get(1).getRepProd(),
							resources,
							this),
							divManAgents.get(0).getProdValue(),
							this
							),												
																			
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(1).getRepProd(), 
							divManAgents.get(0).getProdValue(),
							resources,
							this),
							divManAgents.get(1).getRepProd(),
							this
							),				
							
						discount,
						true);
		 

		double minCompensation1 = calcMinPayment(	divManAgents.get(0).getProdValue(), 
																		divManAgents.get(1).getRepProd() );
		
		relCompensation1 = MathOperations.roundTo2( 
																	(	divManAgents.get(0).getCompensation() 
																		- 
																		minCompensation1		) 
																	/
																	(	maxCompensation1 
																		- 
																		minCompensation1	 ) 
																	);
		
		 System.out.println("\n compensation1:" + divManAgents.get(0).getCompensation()
				 +"\n minCompensation1:" + minCompensation1
				 +"\n maxCompensation1: " + maxCompensation1
				 +"\n relCompensation1: " + relCompensation1);
		 
		
		 double maxCompensation2 = GrovesMechanism.calcCompensation(
					
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(1).getProdValue(), 
							divManAgents.get(0).getRepProd(),
							resources,
							this),
							divManAgents.get(1).getProdValue(),
							this
							),												
																			
					ProfitCalculation.calcProfit(ResourceAllocation.allocate(
							divManAgents.get(0).getRepProd(), 
							divManAgents.get(1).getProdValue(),
							resources,
							this),
							divManAgents.get(0).getRepProd(),
							this
							),				
							
						discount,
						true);
		 
		 double minCompensation2 = calcMinPayment(	divManAgents.get(1).getProdValue(), 
																		divManAgents.get(0).getRepProd() );
						
		
		 relCompensation2 = MathOperations.roundTo2( 	
				 																		( 	divManAgents.get(1).getCompensation() 
				 																			- 
				 																			minCompensation2
				 																		)					
				 																		/
				 																		(
				 																			maxCompensation2
				 																			- 
				 																			minCompensation2
				 																		)
				 																	);
		 
		 System.out.println("\n compensation2:" + divManAgents.get(1).getCompensation()
				 + "\n minCompensation2:" + minCompensation2
				 +"\n maxCompensation2: " + maxCompensation2
				 +"\n relCompensation2: " + relCompensation2);
		
		}

	/* (non-Javadoc)
	 * @see observer.Observer#observeReportingBehavior()
	 */
	public void observeReportingBehavior(){
			delta1= MathOperations.roundTo2(divManAgents.get(0).getRepProd() - divManAgents.get(0).getProdValue());
			relDelta1 = MathOperations.roundTo2(calcRelDelta(0));
			delta2= MathOperations.roundTo2(divManAgents.get(1).getRepProd() - divManAgents.get(1).getProdValue());
			relDelta2 = MathOperations.roundTo2(calcRelDelta(1));
	}
	

	/* (non-Javadoc)
	 * @see observer.Observer#calcTruthfulProfit()
	 */
	public double calcTruthfulProfit() throws IOException{
		
		return 
		ProfitCalculation.calcProfit(ResourceAllocation.allocate(	
																										divManAgents.get(0).getProdValue(), 
																										divManAgents.get(1).getProdValue(),
																										resources,
																										this),
															divManAgents.get(0).getProdValue(),
															this)
		
		+
		
		ProfitCalculation.calcProfit(ResourceAllocation.allocate(	
																										divManAgents.get(1).getProdValue(), 
																										divManAgents.get(0).getProdValue(),
																										resources,
																										this),
																divManAgents.get(1).getProdValue(),
																this);
	}
	
	/* (non-Javadoc)
	 * @see observer.Observer#calcRelDelta(int)
	 */
	public double calcRelDelta(int index){
		if(divManAgents.get(index).getRepProd() < divManAgents.get(index).getProdValue())
			
			return			(divManAgents.get(index).getRepProd()  - divManAgents.get(index).getProdValue())
								/ (divManAgents.get(index).getProdValue() - pMin );
		
		else if(divManAgents.get(index).getRepProd() == divManAgents.get(index).getProdValue())
			
			return  0.0;
		
		else if(divManAgents.get(index).getRepProd() > divManAgents.get(index).getProdValue())
			
			return 			(divManAgents.get(index).getRepProd()  - divManAgents.get(index).getProdValue())
								/ (pMax - divManAgents.get(index).getProdValue() );
		
		return -1111.1;
	}

	/* (non-Javadoc)
	 * @see observer.Observer#printOutput()
	 */
	public void printOutput() throws IOException{
	verificationOutput.append(	 
	//	"\n" + designPoint +
			";" + divManAgents.get(0).getProdValue() + 
			";" + divManAgents.get(0).getRepProd() + 
			";" + delta1 +
			";" + relDelta1+
			";" + divManAgents.get(0).getResourcesAllocated() +
			";" +divManAgents.get(0).getProfit()+
		";" + divManAgents.get(0).getCompensation() +
			";" + relCompensation1+
			
		";" + divManAgents.get(1).getProdValue() +
		";" + divManAgents.get(1).getRepProd() +
		";" + delta2+
			";" + relDelta2 +
			";" + divManAgents.get(1).getResourcesAllocated() +
			";" + divManAgents.get(1).getProfit()+
			";" + divManAgents.get(1).getCompensation() +
			";" + relCompensation2 +
			
		";" + profit+
		";" + calcTruthfulProfit() +
		";" + loss +
		";" + hits +			
			
			";" + resources +
			";" + pMin +
			";" + prodScale +
			";" + optCapacity+
			";" + maxCapacity +
			";" + reductProd +
			";" + discount+
			//";" +
			/*ProfitCalculation.calcProfit(ResourceAllocation.allocate(
					divManAgents.get(1).getProdValue(), 
					divManAgents.get(0).getRepProd(),
					resources,
					this),
					divManAgents.get(1).getProdValue(),
					this
					)+*/
			";"+ // lamda
			";" + //k
			";" + 0 + // 0 for ZI model
			";" + numberOfRuns +
			";" + numberOfTicks
			);
	}
	
}
